#include <alloc.h>
#include <stdio.h>
#include <string.h>
#include "define.h"
#include "tools.h"

struct
{
  char sFromNode[10];
  char sFromUser[10];
  char cFlag;
  char sString[256];
  char sToWho[21];
  unsigned long lDateTime;
} Announce;

//--------------------------------------------------------------------------
//------ Ajouter une annonce dans le fichier d'annonce ---------------------
//--------------------------------------------------------------------------
void ANNOUNCE_add(char * FromNode, char * ToWho, char * FromUser, char Flag, char * String, unsigned long DateTime)
{
  FILE * fPtr;

  //Ouvrir le fichier et placer la nouvelle annonce  la fin
  fPtr = fopen(ANNOUNCE_FILE, "a+b");

  if( ! fPtr )	//Le fichier n'a pas pu tre ouvert
  {
    perror("fopen in ANNOUNCE_add");
    return;
  }/*End IF*/

  //Limiter la longueur des champs
  TOOLS_maxLength(ToWho, 20);
  TOOLS_maxLength(String, 255);

  //Prparer les champs de l'enregistrement
  memset(&Announce, 0, sizeof(Announce));	//Pour rendre le fichier plus clean
  strcpy(Announce.sFromNode, FromNode);
  strcpy(Announce.sToWho,    ToWho);
  strcpy(Announce.sFromUser, FromUser);
  strcpy(Announce.sString,  String);
  Announce.cFlag     = Flag;
  Announce.lDateTime = DateTime;

  //Enregistrer l'annonce  la fin du fichier
  fwrite(&Announce, sizeof(Announce), 1, fPtr);

  fclose(fPtr);
}

//--------------------------------------------------------------------------
//------ Charger une annonce du fichier annonce ----------------------------
//--------------------------------------------------------------------------
//La fonction returne TRUE si l'annonce  pu tre lue
int ANNOUNCE_get(int Pos, char * FromNode, char * ToWho, char * FromUser, char * Flag, char * String, unsigned long * DateTime)
{
  FILE *fPtr;
  long	lOffset;

  //Ouvrir le fichier
  fPtr = fopen(ANNOUNCE_FILE, "rb");

  if( ! fPtr )	//Le fichier n'a pas pu tre ouvert
    return FALSE;

  //Se placer  la fin du fichier
  fseek(fPtr, 0L, SEEK_END);

  //Remonter de Pos enregistrements
  lOffset = (long) Pos * (long) sizeof(Announce);
  if( lOffset > ftell(fPtr) )
  {
    fclose( fPtr );
    return FALSE;
  }
  fseek(fPtr, -lOffset, SEEK_CUR);

  //Extraire les champs
  fread(&Announce, sizeof(Announce), 1, fPtr);

  //Passer les champs aux variables
  strcpy(FromNode, Announce.sFromNode);
  strcpy(ToWho,    Announce.sToWho);
  strcpy(FromUser, Announce.sFromUser);
  strcpy(String,   Announce.sString);
  *Flag     = Announce.cFlag;
  *DateTime = Announce.lDateTime;

  //Fermer le fichier et retourner
  fclose(fPtr);
  return TRUE;
}

/*---------------------------------------------------------------------------
  ------ Rechercher si l'annonce est un double ------------------------------
  ---------------------------------------------------------------------------
  La fonction retourne TRUE si l'annonce est double */
int ANNOUNCE_isDupe(char * FromNode, char * ToWho, char * FromUser, char * String, unsigned long DateTime)
{
	#define	NLAST	50L
	FILE   *fPtr;
	long	lOffset;
	int	index;
	char	szLogger[16];

	/* Recherche de DUPE a effectuer qqs le SSID du logger */
	strcpy(szLogger, FromUser);
	TOOLS_removeSsid(szLogger);

	/* Ouvrir le fichier */
	fPtr = fopen(ANNOUNCE_FILE, "rb");

	if( ! fPtr )	/* Impossible d'ouvrir */
	{
		perror("fopen in ANNOUNCE_isDupe");
		return FALSE;
	}

	/* Placer le pointeur de fichier NLAST spots en arriere */
	fseek(fPtr, 0L, SEEK_END);			/* Fin du fichier */
	lOffset =  NLAST * (long) sizeof(Announce);	/* Calcul offset */
	if( lOffset > ftell(fPtr) )
		fseek(fPtr, 0L, SEEK_SET);		/* Moins de NLAST spots enregistres : se placer en debut de fichier */
	else
		fseek(fPtr, -lOffset, SEEK_CUR);	/* Pointer NLAST */

	/* Tester chacune des annonces */
	for(index = 0; index < NLAST; index++)
	{
		/* Lecture enregistrement */
		if( ! fread(&Announce, sizeof(Announce), 1, fPtr) )
			break;	/* EOF */

		/* Comparer ... */
		TOOLS_removeSsid(Announce.sFromUser);	/* Qqs le SSID du logger */
		if( ! strcmp(FromNode, Announce.sFromNode) &&
		    ! strcmp(ToWho, Announce.sToWho)       &&
		    ! strcmp(szLogger, Announce.sFromUser) &&
		    ! strcmp(String, Announce.sString) )
		{
			/* Fenetre de 2 min */
			if( DateTime > (Announce.lDateTime + 300L) )
				continue;

			/* DUPE */
			fclose( fPtr );
			return TRUE;
		}

	}/* End FOR */

	/* Fermer le fichier et retourner */
	fclose( fPtr );
	return FALSE;
}