#include <alloc.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <dos.h>
#include "buffers.h"
#include "define.h"
#include "protocol.h"
#include "clulink.h"
#include "users.h"
#include "set.h"
#include "log.h"
#include "switch.h"


/*

- BUFFERS 1  64 : les changes avec le SWITCH
- BUFFER  0      : fentre F1 uniquement
- BUFFER  65     : fentre F2 uniquement
*/

//--------------------------------------------------------------------------
//------ Variables globales ------------------------------------------------
//--------------------------------------------------------------------------
Channel Stream[MAX_BUFFER+1]; 		//Rserver 65 Structures Channel

extern	char	STREAMS_level[MAX_STREAMS];
extern  char	PROTOCOL_linkType[65];

extern  tUserCfg * pUserCfg;

#ifdef DEBUG_LOG_
extern	char  STREAMS_callsign[MAX_STREAMS][10];
extern  char *PARAMS_logs;
#endif

//-------------------------------------------------------------------------
//------ Initialiser les buffers ------------------------------------------
//-------------------------------------------------------------------------
void BUFFERS_init(void)
{
  int index;

  for(index = 0; index <= MAX_BUFFER; index++)
  {
    Stream[index].pBuffer[IN]  = NULL;
    Stream[index].pBuffer[OUT] = NULL;
    Stream[index].iSize[IN]    = NULL;
    Stream[index].iSize[OUT]   = NULL;
  }
}

//--------------------------------------------------------------------------
//------ openbuff - ouvre un nouveau buffer pour le stream N "streamnum" --
//------ et le charge avec une nouvelle chaine de caracteres ---------------
//--------------------------------------------------------------------------
int BUFFERS_openBuff(int StreamNum, char * String, int Direction, int StrSize)
//Rserve de la mmoire tampon
//Retourne DONE  si OK
//Retourne ERR_NOT_ENOUGH_MEMORY si la mmoire disponible est insuffisante
{

  Stream[StreamNum].pBuffer[Direction] = (char far*) farmalloc( (unsigned long) StrSize + 1L );

  //Le buffer a t-il pu tre ouvert ?
  if( ! Stream[StreamNum].pBuffer[Direction] )
    return ERR_NOT_ENOUGH_MEMORY;

  memcpy(Stream[StreamNum].pBuffer[Direction], String, StrSize);
  Stream[StreamNum].pBuffer[Direction][StrSize] = SNULL;
  Stream[StreamNum].iSize[Direction] = StrSize;

  return DONE;
}

//--------------------------------------------------------------------------
//------ addbuff - ajoute des donnes de longueur StrSize dans un buffer ---
//--------------------------------------------------------------------------
int BUFFERS_addBuff_sized(int StreamNum, char * String, int Direction, int StrSize)
//Agrandit la mmoire Tampon
//Empile le nouveau string
//Retourne DONE si OK
//Retourne ERR_NOT_ENOUGH_MEMORY si la mmoire disponible est insuffisante
{
  char far * pPtr;
  int	     iLastSize;

  if( StreamNum >= RCMD_VIRTSTREAM )
  {
	/* C'est pour un remote cmd - ecrire le contenu du buffer dans
	   le fichier tmp */
	FILE *	fPtr;

	fPtr = fopen(TEMP_RCMD_FILE, "a+t");
	if( ! fPtr )
	{
		/* Erreur d'ouverture du fichier */
		perror("fopen in BUFFERS_addBuff_sized (RCMD)");
		return FALSE;
	}

	fwrite(String, StrSize, 1, fPtr);
	fclose(fPtr);
	return TRUE;
  }

  //Eviter les erreurs de bords
  if( StreamNum < 0 || StreamNum > 65 )
    return FALSE;

  if( ! Stream[StreamNum].pBuffer[Direction] ) //Buffer pas encore ouvert
  {
    if( BUFFERS_openBuff(StreamNum, String, Direction, StrSize) == DONE )
      return DONE;
    else
      return ERR_NOT_ENOUGH_MEMORY;
  }

  //------ Calcule la nouvelle longueur des chaines ------
  iLastSize = Stream[StreamNum].iSize[Direction];
  Stream[StreamNum].iSize[Direction] += StrSize;

#ifndef LINUX
  /* Ne pas laisser farrealloc utiliser toute la memoire */
  if( coreleft() < 7500L )
	return ERR_NOT_ENOUGH_MEMORY;
#endif

  //Rallouer un bloc
  pPtr = (char far *) farrealloc( Stream[StreamNum].pBuffer[Direction], (unsigned long) Stream[StreamNum].iSize[Direction] + 1L);

  //------ Si la rallocation  choue, addbuff retourne la valeur NULL ---
  //------ et l'information  bufferiser est perdue ------------------------
  if( ! pPtr )
    return ERR_NOT_ENOUGH_MEMORY;

  Stream[StreamNum].pBuffer[Direction] = pPtr;

  //Copie du bloc au nouvel emplacement
  memcpy(Stream[StreamNum].pBuffer[Direction]+iLastSize, String, StrSize);
  Stream[StreamNum].pBuffer[Direction][iLastSize + StrSize] = SNULL;

  return DONE;
}

//---------------------------------------------------------------------------
//------ getRow - lit une ligne COMPLETE dans le buffer ---------------------
//---------------------------------------------------------------------------
int BUFFERS_getRow(int StreamNum, char * String, int Direction)
//Retourne 0 si le buffer est vide
//Retourne la longueur de la chaine sinon.
//La ligne est retourne sans le caractre de RC
{
  char far * 	pStrOrg;
  char far * 	pStrPos;
  unsigned int  index;
  unsigned int  iLen;
  int		iCluLinkFrame = FALSE;
  int		iStrSize = 0;		//Longueur de la chaine retourne

  //Initialisation de String
  String[0] = SNULL;

  //Est-ce une trame CluLink ?
  if( (STREAMS_level[StreamNum] & LEVEL_cluster) && (PROTOCOL_linkType[StreamNum] == CLU_LINK) )
    iCluLinkFrame = TRUE;

  //------ Le buffer est-il ouvert ? ---------------------------------------
  if( ! Stream[StreamNum].pBuffer[Direction] )
  {
    String[0] = SNULL;
    return 0;		//Retourner FALSE (buffer vide)
  }

  if( iCluLinkFrame )
  {
    //------ Le premier caractre indique la longueur du protocole ----------
    //------ Ne pas lire si l'intgralit de ce protocole n'a pas -----------
    //------ t reue ------------------------------------------------------
    if( Stream[StreamNum].iSize[Direction] < CLULINK_getSize(Stream[StreamNum].pBuffer[Direction]) )
      return FALSE;

    //Dfinir le nombre de caractres  copier
    iLen = CLULINK_getSize(Stream[StreamNum].pBuffer[Direction]) + 2;
  }
  else
  {
    //------ Ne pas lire le buffer tant que la fin d'une ligne n'a pas -------
    //------ t reue -------------------------------------------------------
    if( ! strchr( Stream[StreamNum].pBuffer[Direction], '\r') )
      return FALSE;

    //Dfinir le nombre de caractres  copier (au maximum)
    iLen = 255;
  }

  //------ Copier la ligne dans String ---------------------------------------
  pStrOrg = Stream[StreamNum].pBuffer[Direction];
  pStrPos = String;

  for(index = 0; index < iLen && Stream[StreamNum].iSize[Direction] != 0; index++)
  {
    *pStrPos = *pStrOrg;
    Stream[StreamNum].iSize[Direction]--;
    iStrSize++;


    if( *pStrPos || iCluLinkFrame )
      pStrOrg++;

    if( (*pStrPos == SNULL || *pStrPos == '\r') && ! iCluLinkFrame )
    {
      *pStrPos = SNULL;	//Ne pas retourner le caractre \r
      break;
    }
    else
      pStrPos++;
  }

  //------ Ajouter un caractre NULL  la fin de String, si ncessaire -----
  if( *pStrPos != SNULL )
    *++pStrPos = SNULL;

#ifdef DEBUG_LOG_
  if( SET_check(PARAMS_logs, LOG_BUFFER_STR) )
    LOG_write(LOG_BUFFEROUT_FILE, LOG_DATE_YES, "getRow  [%02d] %c%s %s\n", StreamNum, (Direction == IN ? '<' : '>'), STREAMS_callsign[StreamNum], String);
#endif

  //------ La fin du buffer est-elle atteinte ? ----------------------------
  if( Stream[StreamNum].iSize[Direction] == 0 )
  {
    farfree( Stream[StreamNum].pBuffer[Direction] );
    Stream[StreamNum].pBuffer[Direction] = NULL;
    return iStrSize;
  }

  //------ Sinon, rallouer le bloc avec la chaine restante dans le buffer -
  memmove(Stream[StreamNum].pBuffer[Direction], pStrOrg, Stream[StreamNum].iSize[Direction]);
  Stream[StreamNum].pBuffer[Direction][Stream[StreamNum].iSize[Direction]] = SNULL;
  Stream[StreamNum].pBuffer[Direction] = (char far *) farrealloc(Stream[StreamNum].pBuffer[Direction], (unsigned long) Stream[StreamNum].iSize[Direction] + 1L);
  return iStrSize;
}

//---------------------------------------------------------------------------
//------ getBloc - lit un bloc dans le buffer -------------------------------
//---------------------------------------------------------------------------
//Retourne 0 si le buffer est vide;
//Retourne la longueur de la chaine sinon.
int BUFFERS_getBloc(int StreamNum, char * String, int Direction)
{
  char far * 	pStrOrg;
  char far * 	pStrPos;
  unsigned int  index;
  int		iCluLinkFrame = FALSE;
  int		iStrSize = 0;	//Longueur de la chaine retourne
  unsigned int	iPaclen = 255;
  
  //Initialisation de String
  String[0] = SNULL;

  //Est-ce une trame CluLink ?
  if( STREAMS_level[StreamNum] & LEVEL_cluster && PROTOCOL_linkType[StreamNum] == CLU_LINK )
    iCluLinkFrame = TRUE;

  //------ Le buffer est-il ouvert ? ---------------------------------------
  if( ! Stream[StreamNum].pBuffer[Direction] )
  {
    String[0] = SNULL;
    return 0;		//Retourne 0 (buffer vide)
  }

#ifdef LINUX  
  //Gestion du paclen - pour linux uniquement
  if( Direction == OUT && StreamNum >= 1 && StreamNum <= 64 )
    iPaclen = SWITCH_getPaclen(StreamNum);
#endif

  //------ Copier la ligne dans String SANS jamais depasser 255 caractres -
  pStrOrg = Stream[StreamNum].pBuffer[Direction];
  pStrPos = String;

  for(index = 0; index < iPaclen && Stream[StreamNum].iSize[Direction] != 0; index++)
  {
    *pStrPos = *pStrOrg;
    Stream[StreamNum].iSize[Direction]--;
    iStrSize++;

    if( *pStrPos || iCluLinkFrame )
      pStrOrg++;

    if( ((*pStrPos == SNULL || *pStrPos == '\r') && ! iCluLinkFrame) || Stream[StreamNum].iSize[Direction] == 0 )
      break;
    else
      pStrPos++;
  }

  //------ Ajouter un caractre NULL  la fin de String, si ncessaire -----
//  if( index == 255 )	//Puisque pStrPos avait incrment inutilement
//    pStrPos--;
  String[iStrSize] = SNULL;
  if( *pStrPos != SNULL )
    *++pStrPos = SNULL;	//J'ai enlev le ++

#ifdef DEBUG_LOG_
  if( SET_check(PARAMS_logs, LOG_BUFFER_STR) )
    LOG_write(LOG_BUFFEROUT_FILE, LOG_DATE_YES, "getBloc [%02d] %c%s %s\n", StreamNum, (Direction == IN ? '<' : '>'), STREAMS_callsign[StreamNum], String);
#endif

  //------ La fin du buffer est-elle atteinte ? ----------------------------
  if( /**pStrOrg == NULL*/ Stream[StreamNum].iSize[Direction] == 0 )
  {
    farfree( Stream[StreamNum].pBuffer[Direction] );
    Stream[StreamNum].pBuffer[Direction] = NULL;
    return iStrSize;
  }

  //------ Sinon, rallouer le bloc avec la chaine restante dans le buffer -
  memmove(Stream[StreamNum].pBuffer[Direction], pStrOrg, Stream[StreamNum].iSize[Direction]);
  Stream[StreamNum].pBuffer[Direction] = (char far *) farrealloc(Stream[StreamNum].pBuffer[Direction], (unsigned long) Stream[StreamNum].iSize[Direction] + 1L);

  return iStrSize;
}

//--------------------------------------------------------------------------
//------ printfbuff - ajoute des donnes dans un buffer --------------------
//--------------------------------------------------------------------------
int BUFFERS_printBuff(int StreamNum, int Direction, char * String, ...)
{
  va_list     args;
  static char buf[256];
  int ReturnValue;

  va_start(args, String);

  vsprintf(buf, String, args);
  ReturnValue = BUFFERS_addBuff(StreamNum, buf, Direction);

  va_end(args);

  return ReturnValue;
}

//--------------------------------------------------------------------------
//------ addbuff - ajoute des donnes dans un buffer -----------------------
//--------------------------------------------------------------------------
int BUFFERS_addBuff(int StreamNum, char * String, int Direction)
{
  return BUFFERS_addBuff_sized(StreamNum, String, Direction, strlen(String));
}

//--------------------------------------------------------------------------
//------ Fermer un buffer --------------------------------------------------
//--------------------------------------------------------------------------
void BUFFERS_close(int StreamNum)
{
  int Direction;

  for(Direction = 0; Direction < 2; Direction++)
  {
    if( Stream[StreamNum].pBuffer[Direction] )
    {
      farfree( Stream[StreamNum].pBuffer[Direction] );
      Stream[StreamNum].pBuffer[Direction] = NULL;
      Stream[StreamNum].iSize[Direction]   = 0;
    }
  }/*End FOR*/
}

/*************************************************************************
 * Les fonctions qui suivent sont utilisees pour gerrer des buffers      *
 * temporaires (pour les commandes UPDATE des databases par exemples)    *
 *************************************************************************/
//---------------------------------------------------------------------------
//------ Ajouter une chaine dans un buffer temporaire -----------------------
//---------------------------------------------------------------------------
void BUFFERS_addTemp(int StreamNum, char * String, int Size)
{
  int	iOldBloc;
  int	iNewBloc;

  //Acceder a la config du stream
  USERS_getRecord(StreamNum);

  //Au premier appel, creer un nouveau buffer ...
  if( pUserCfg->iBufferLength == 0 )
  {
    //Un buffer etait-il deja alloue ? Si oui, le virrer ...
    if( pUserCfg->pBuffer )
      free( pUserCfg->pBuffer );

    pUserCfg->pBuffer = (char *) malloc(1024);
  }
  else
  {
    //Faut-il augmenter la taille du buffer ?
    //Calcul du nombre de blocs necessaires pour l'ancien buffer et le nouveau
    iOldBloc = pUserCfg->iBufferLength / 1024;
    iNewBloc = (pUserCfg->iBufferLength + Size) / 1024;

    //Faut-il augmenter la taille du buffer ...
    if( iOldBloc != iNewBloc )
      pUserCfg->pBuffer = (char *) realloc(pUserCfg->pBuffer, (iNewBloc + 1) * 1024);
  }

  //Copier le nouveau bloc
  memmove(pUserCfg->pBuffer + pUserCfg->iBufferLength, String, Size);

  //Mettre a jour et sauvegarder la config utilisateur
  pUserCfg->iBufferLength += Size;
  USERS_updateLocalConfig(StreamNum);
}

//---------------------------------------------------------------------------
//------ Ajouter une chaine dans un buffer temporaire -----------------------
//---------------------------------------------------------------------------
//La fonction retourne une ligne lue dans le buffer et retourne la longueur
//de la chaine. Si le buffer est vide, la fonction retourne 0.
//Si le buffer n'est pas vide, un caractere terminal NULL est ajoute a
//la fin de la chaine retournee.
int BUFFERS_getTemp(int StreamNum, char * String, int MaxSize)
{
  int 	iSize = 0;
  char *pScr;
  char *pDest;

  //Acceder a la config du stream
  USERS_getRecord(StreamNum);

  if( pUserCfg->iBufferLength == 0 )
    return 0; 				//Buffer vide

  //Pointer sur le debut du buffer
  pScr  = pUserCfg->pBuffer;
  pDest = String;

  //Le fonction ne retourne jamais de chaine suprieure a 255 octets
  for(iSize = 0; iSize < MaxSize; iSize++)
  {
    *pDest = *pScr;

    pUserCfg->iBufferLength--;

    //Fin de la ligne atteinte ??
    if( *pScr == '\r' || pUserCfg->iBufferLength == 0 )
      break;

    pScr++;
    pDest++;

   }/*End FOR*/

   //Mettre a jour et sauvegarder la config utilisateur
   USERS_updateLocalConfig(StreamNum);

   //Deplacer le reste du buffer au debut
   memmove(pUserCfg->pBuffer, ++pScr, pUserCfg->iBufferLength);

   //Ajouter un NULL a la chaine lue et retourner sa longueur
   *++pDest = SNULL;
   return iSize;
}

//---------------------------------------------------------------------------
//------ Fermer (killer) un buffer temporaire -------------------------------
//---------------------------------------------------------------------------
void BUFFERS_killTemp(int StreamNum)
{
  //Acceder a la config du stream
  USERS_getRecord(StreamNum);

  if( pUserCfg->pBuffer )
    free( pUserCfg->pBuffer );

  pUserCfg->pBuffer       = NULL;
  pUserCfg->iBufferLength = 0;

  //Mettre a jour et sauvegarder la config utilisateur
  USERS_updateLocalConfig(StreamNum);
}

//---------------------------------------------------------------------------
//------ Copier un buffer dans un fichier et fermer le buffer ---------------
//---------------------------------------------------------------------------
//La fonction retourne TRUE si OK et FALSE sinon
int BUFFERS_saveTemp(int StreamNum)
{
  FILE *fPtr;
  int	iLength;
  char	sBuffer[256];

  //Acceder a la config du stream
  USERS_getRecord(StreamNum);

  //ouvrir le fichier
  fPtr = fopen(pUserCfg->sFileName, "wt");

  if( ! fPtr )
  {
    perror("fopen in BUFFERS_saveTemp");
    return FALSE;
  }

  //Copier dans le fichier
  for(;;)
  {
    iLength = BUFFERS_getTemp(StreamNum, sBuffer, 255);

    if( iLength == 0 )
      break;		//Fin du buffer atteinte

    //Enregistrer la ligne dans le fichier
    fwrite(sBuffer, iLength, 1, fPtr);
  }/*End LOOP*/

  //Fermer le fichier
  fclose (fPtr);

  //Killer le buffer
  BUFFERS_killTemp(StreamNum);

  return TRUE;		//Ok
}
